/* *********************************************************************
 * Class MetaResolvingService
 * 
 * Copyright (c) 2010-2012, German National Library/Deutsche Nationalbibliothek
 * Adickesallee 1, D-60322 Frankfurt am Main, Federal Republic of Germany 
 *
 * This program is free software.
 * For your convenience it is dual licensed.
 * You can redistribute it and/or modify it under the terms of
 * one of the following licenses:
 * 
 * 1.)
 * The GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * You should have received a copy of the GNU General Public License
 * along with this program (gpl-3.0.txt); if not please read
 * http://www.gnu.org/licenses/gpl.html
 * 
 * 2.)
 * The European Union Public Licence as published by
 * The European Commission (executive body of the European Union);
 * either version 1.1 of the License, or (at your option) any later version.
 * You should have received a copy of the European Union Public Licence
 * along with this program (eupl_v1.1_en.pdf); if not please read
 * http://www.osor.eu/eupl
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the above licenses for more details.
 * 
 * Jürgen Kett, Kadir Karaca Kocer -- German National Library
 * 
 **********************************************************************/

/* ********************************************************************
 * CHANGELOG:
 * 
 * 2012-04-03 Port to Maven by Timo Heck & Karaca Kocer
 * 2010-06-11 Code formatting and license text by Karaca Kocer
 * 2010-05-05 Refactoring + new package structure by Karaca Kocer
 * Created on 2010-03-17 by Jürgen Kett
 ********************************************************************/

package org.culturegraph.resolver.api;

import java.util.List;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import eu.europeanaconnect.erds.DataProvider;
import eu.europeanaconnect.erds.ResolverException;
import eu.europeanaconnect.erds.ResolverException.ResolverExceptionCode;
import eu.europeanaconnect.erds.ResolverRequest;
import eu.europeanaconnect.erds.ResolverResponse;
import eu.europeanaconnect.erds.ResourceDescription;

/**
 * base service for the metaresolver.
 */
public class MetaResolvingService extends eu.europeanaconnect.erds.ResolvingService {
    private static final Log LOGGER = LogFactory.getLog(MetaResolvingService.class);
	private List<ResourceDescriptionProvider> resourceDescriptionProviders;

	/**
	 * @return List of the providers that deliver bibliographic meta data.
	 */
	public List<ResourceDescriptionProvider> getResourceDescriptionProviders() {
		return this.resourceDescriptionProviders;
	}

	/**
	 * @param providers the metaresolver uses
	 * @since 17.03.2010
	 */
	public void setResourceDescriptionProviders(List<ResourceDescriptionProvider> providers) {
		this.resourceDescriptionProviders = providers;
	}

	/**
	 *{@inheritDoc}
	 */
	@Override
	public ResolverResponse getResponse(ResolverRequest resolverRequest) throws ResolverException {
		ResolverResponse ret = null;
		ResolverException storedException = null;
		
		if (LOGGER.isDebugEnabled()) LOGGER.debug("MetaResolvingService.getResponse(): Looking for the right resolver");
		// resolve identifier
		boolean resolverFound = false;
		for (DataProvider<ResolverResponse> resolver : this.resolvers) {
			if (resolverRequest.getIdentifier().toLowerCase().matches(resolver.getIdentifierPattern())) {
				resolverFound = true;
				try {
					if (LOGGER.isDebugEnabled()) LOGGER.debug("MetaResolvingService.getResponse(): Found a matching resolver. Trying to get a response.");
					ret = resolver.getResponse(resolverRequest);
					if (ret != null) {
						if (LOGGER.isDebugEnabled()) LOGGER.debug("MetaResolvingService.getResponse(): Got a response. Analyzing.");
						ret.setDataProviderId(resolver.getId());
						storedException = null;
						//TODO: check this
						if (ret.getUrl() != null) {
							if (LOGGER.isDebugEnabled()) LOGGER.debug("MetaResolvingService.getResponse(): URL found. Returning.");
							break;
						}
					} else {
						if (LOGGER.isDebugEnabled()) LOGGER.debug("MetaResolvingService.getResponse(): Response was NULL!");
					}
				} catch (ResolverException e) {
					LOGGER.warn("MetaResolvingService.getResponse(): ResolverException!");
					storedException = e;
				} catch (RuntimeException e) {
					LOGGER.error("MetaResolvingService.getResponse(): RuntimeException! Message is:\n" + e.getLocalizedMessage());
					throw new ResolverException(this.id, ResolverExceptionCode.UNKNOWN_ERROR);
				} catch (Exception e) {
					LOGGER.error("MetaResolvingService.getResponse(): Exception! Message is:\n" + e.getLocalizedMessage());
					throw new ResolverException(this.id, ResolverExceptionCode.UNKNOWN_ERROR);
				}
			}
		}
		
		// TODO: check if this works (if a resolver fails, the algorithms checks the next one)
        if (storedException != null) {
        	LOGGER.warn("MetaResolvingService.getResponse(): None of the registered resolvers match the pattern");
            throw storedException;
        }
        
		// try to fetch resource description (optional)
        //TODO: refactor to make more than one resource descriptions possible
		if ((this.resourceDescriptionProviders != null) && (ret != null) && (ret.getResourceDescription()) == null) {
			if (LOGGER.isDebugEnabled()) LOGGER.debug("Fetching the resource description");
			for (ResourceDescriptionProvider resourceDescriptionProvider : this.resourceDescriptionProviders) {
				if (resolverRequest.getIdentifier().toLowerCase().matches(
						resourceDescriptionProvider.getIdentifierPattern())) {
					try {
						ResourceDescription resourceDescription = resourceDescriptionProvider.getResponse(resolverRequest);
						resourceDescription.setDataProviderId(resourceDescriptionProvider.getId());
						ret.setResourceDescription(resourceDescription);
						break;
					} catch (ResolverException e) {
						//ignore the exception. try the next provider.
					}
				}
			}
		}
		
		if (ret == null) {
			if (resolverFound) {
				//resolver is found but the identifier is still unknown
				throw new ResolverException(this.id, ResolverExceptionCode.IDENTIFIER_NOT_REGISTERED);
			}
			throw new ResolverException(this.id, ResolverExceptionCode.UNSUPPORTED_IDENTIFIER_SCHEME);
		}
		
		return ret;
	}
}
