/**********************************************************************
 * Class AbstractController
 *  
 * Copyright (c) 2011-2012, German National Library/Deutsche Nationalbibliothek
 * Adickesallee 1, D-60322 Frankfurt am Main, Federal Republic of Germany 
 *
 * This program is free software.
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 * Thomas Haidlas -- German National Library
 * Kadir Karaca Kocer -- German National Library
 **********************************************************************/

/* ********************************************************************
 * CHANGELOG:
 * 
 * 2013-03-22 HTTP Response Codes implemented
 * 2012-04-03 Port to Maven by Timo Heck & Karaca Kocer
 * 2011-12-12 License information and JavaDocs by Karaca Kocer
 * Created on 2011-12-08 by Thomas Haidlas -- German National Library
 ********************************************************************/

package org.nbnResolving.resolver.controller.impl;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.culturegraph.resolver.api.MetaResolvingService;
import org.nbnResolving.pidef.HeaderType;
import org.nbnResolving.pidef.MessageType;
import org.nbnResolving.pidef.PidefDocument;
import org.nbnResolving.resolver.Constants;
import org.nbnResolving.resolver.PidefDocumentGenerator;
import org.nbnResolving.resolver.controller.Controller;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.web.servlet.ModelAndView;

import eu.europeanaconnect.erds.ResolverException;
import eu.europeanaconnect.erds.ResolverRequest;
import eu.europeanaconnect.erds.ResolverResponse;

/**
 * An abstract controller for resolving.
 * 
 * @author Thomas Haidlas
 * @author K. Karaca Kocer
 */
public abstract class AbstractController implements Controller {
	
    private static final Log LOGGER = LogFactory.getLog(AbstractController.class);
	
	@Autowired
	protected MetaResolvingService metaResolver;
	private String viewName;
	
	/** @return Name of the View */
	public String getViewName() {
		return this.viewName;
	}

	protected void setViewName(String newViewName) {
		this.viewName = newViewName;
	}

	/**
	 * @return The service used
	 * @see org.culturegraph.resolver.api.MetaResolvingService
	 * */
	public MetaResolvingService getMetaResolver() {
		return this.metaResolver;
	}

	/**
	 * @param resolver Set the service used
     * @see org.culturegraph.resolver.api.MetaResolvingService
     * */
	public void setMetaResolver(MetaResolvingService resolver) {
		this.metaResolver = resolver;
	}

	/**
	 * @param resolverRequest What the user requested.
	 * @param req HTTP request.
	 * @param res HTTP response.
	 * @return A ModelAndView object to render.
	 */
	public ModelAndView resolveId(final HttpServletRequest req, final HttpServletResponse res, final ResolverRequest resolverRequest) {
		final PidefDocument pidefDocument = getResponseDoc(resolverRequest, req);
		res.setStatus(determineResponseCode(pidefDocument));
		return getModelAndView(pidefDocument);
	}
	   
    /**
     * @param resolverRequest What the user requested.
     * @param req HTTP request.
     * @param res 
     * @param vName Name of the View
     * @return A ModelAndView object to render.
     */
	public ModelAndView resolveId(final HttpServletRequest req, final HttpServletResponse res,
	                              final ResolverRequest resolverRequest, final String vName) {
		final PidefDocument pidefDocument = getResponseDoc(resolverRequest, req);
        res.setStatus(determineResponseCode(pidefDocument));
		return getModelAndView(pidefDocument, vName);
	}
   
    /**
     * @param identifier The unique identifier to resolve.
     * @param req HTTP request.
     * @param res 
     * @param vName Name of the View
     * @return A ModelAndView object to render.
     */
	public ModelAndView resolveId(final HttpServletRequest req, final HttpServletResponse res, 
	                              final String identifier, final String vName) {
		ResolverRequest resolverRequest = new ResolverRequest();
		resolverRequest.setAction("full");
		resolverRequest.setIdentifier(identifier);
		return resolveId(req, res, resolverRequest, vName);
	}
	
	@Override
    public ModelAndView resolveId(final HttpServletRequest req, final HttpServletResponse res, final String identifier) {		
		ResolverRequest resolverRequest = new ResolverRequest();
		resolverRequest.setAction("full");
		resolverRequest.setIdentifier(identifier);
		return resolveId(req, res, resolverRequest);
	}
		
	private PidefDocument getResponseDoc(final ResolverRequest resolverRequest, final HttpServletRequest req){
		ResolverResponse resolverResponse = null;
		try {
			resolverResponse = this.metaResolver.getResponse(resolverRequest);
		} catch (ResolverException e) {
			LOGGER.info("getResponseDoc(): no response for request: " + convertUserRequest(req) + " returning an error header.");
			/*XmlOptions options = new XmlOptions();
			options.setCharacterEncoding("UTF-8");
			options.setSaveNamespacesFirst();
			options.setValidateStrict();
			PidefDocument doc = PidefDocument.Factory.newInstance(options);
			HeaderType header = doc.addNewPidef().addNewHeader();
			*/
			PidefDocument doc = PidefDocumentGenerator.generateNewPidefDocument();
			HeaderType header = doc.getPidef().getHeader();
			header.setStatusCode(Constants.ERROR_DURING_RESOLVING);
			header.setSource("Resolver GUI");
			MessageType message = header.insertNewMessage(0);
			message.setLang("en");
			message.setStringValue(e.getMessage());
			header.setRequest(convertUserRequest(req));
			e.printStackTrace();
			return doc;
		}

		org.nbnResolving.resolver.EpicurResolverResponse rr = null;
		if (resolverResponse instanceof org.nbnResolving.resolver.EpicurResolverResponse) {
			LOGGER.debug("getResponseDoc(): org.nbnResolving.resolver.EpicurResolverResponse found. Casting.");
			rr = (org.nbnResolving.resolver.EpicurResolverResponse) resolverResponse;
		} else {
			rr = new org.nbnResolving.resolver.EpicurResolverResponse();
			//rr.getDoc().getPidef().getData().addNewResolvingInformation().addNewPiInfo().setIdentifier(pi);
			if (resolverResponse.getUrl() != null) {
				LOGGER.debug("getResponseDoc(): Answer of a remote institution. Writing the URL in PIDEF");
				rr.getDoc().getPidef().getHeader().setStatusCode(Constants.REMOTE_RESOLVER_INFORMATION);
				rr.getDoc().getPidef().getHeader().setSource(resolverResponse.getDataProviderId());
				rr.getDoc().getPidef().getData().addNewLinks().addLink(resolverResponse.getUrl());
				//rr.getDoc().getPidef().getData().getLinks().getLinkArray(0);
			} else {
				LOGGER.debug("getResponseDoc(): Answer of a remote institution. No URL found!");
				rr.getDoc().getPidef().getHeader().setStatusCode(Constants.REMOTE_RESOLVER_INFORMATION);
				rr.getDoc().getPidef().getHeader().setSource(resolverResponse.getDataProviderId());
				//TODO: error handling! two separete status codes are needed
			}
		}
		rr.getDoc().getPidef().getHeader().setRequest(convertUserRequest(req));
		
		return rr.getDoc();
	}
	
	private ModelAndView getModelAndView(final PidefDocument pidefDocument) {
		return getModelAndView(pidefDocument, this.viewName);
	}
	
    private ModelAndView getModelAndView(final PidefDocument pidefDocument, final String vName) {
        ModelAndView ret = new ModelAndView();
        ret.addObject("doc", pidefDocument);
        if ((vName != null) && (!vName.isEmpty())) {
            ret.setViewName(vName);
        } else {
            ret.setViewName(this.viewName); 
        }
        return ret;
    }
    
	private static String convertUserRequest(final HttpServletRequest req) {
		String url   = req.getRequestURL().toString();
		String query = req.getQueryString();	
		if ((query != null) && !query.isEmpty()) {
			url = url + "?" + query;
		}
		return url;
	}
    
    private static int determineResponseCode(final PidefDocument doc) {
        int code = doc.getPidef().getHeader().getStatusCode();
        switch (code) {
            case Constants.NO_SUCH_URN_IN_DATABASE:                
            case Constants.ERROR_DURING_RESOLVING:                
            case Constants.NO_URN_WITH_THIS_URL:
            case Constants.NO_NAMESPACE_FOUND:
            case Constants.NO_INSTITUTION_FOUND:
                return HttpServletResponse.SC_NOT_FOUND;
                
            case Constants.URN_IS_NULL:
            case Constants.WRONG_URN_FORMAT:
                return HttpServletResponse.SC_BAD_REQUEST;

            case Constants.REQUESTED_URN_IS_INACTIVE:
                return HttpServletResponse.SC_GONE;
                
            case Constants.NO_SUCH_URN_STATUS:
            case Constants.PROBLEM_WITH_URN_ID:
                return HttpServletResponse.SC_INTERNAL_SERVER_ERROR;

            default:
                return HttpServletResponse.SC_OK;
        }
    }
}
